/** Representation of a cubic slice through the space 
represented by the minterms. Although called a "Cube" it
can be any number of dimensions, from a minterm which is
a single dimensionless point through a 6 dimension 
hypercube.
*/
class Cube implements Cloneable{
	protected long map;  // Set of points in the cube
	protected int mustBeZero, mustBeOne; /* Coordinate
	      based description of the cube. If a point has
	      0 in all bits matching mustBeZero and one in 
	      each bit matching mustBeOne it is in the cube.*/
	int dims;  /* Current number of dimensions occupied by
	      the cube.*/
	      
	/** Create a cube representing a single point.
	@param point The minterm number of the single point that 
	forms the initial cube.
	*/
	Cube(int point){
		mustBeZero = ~point;
		mustBeOne = point;
		map = 1L << point;
    //System.out.println("New cube "+toString()+" point "+point);
	}
	/** Expand a cube by sweeping it in a specified direction.
	@param dim The index of the dimension in which expansion is
	proposed.
  @param permittedPoints The set of points that can be in the
  cube.
	@return A new cube that contains the expanded version. Null
	if any point that would be added in the expansion is not in
	the permittedPoints set from creation of this cube.
	@throw SimplifierError attempt to expand the cube in a direction
	in which is already has depth.
	*/
	Cube expand(int dim,long permittedPoints)throws SimplifierError{
		int mask = 1 << dim;
		if((mustBeZero & mask) != 0)
			return expandUp(mask,permittedPoints);
		else if((mustBeOne & mask) != 0)
			return expandDown(mask,permittedPoints);
		else
			throw new SimplifierError("Duplicate expansion");
	}
  protected Cube expandUp(int mask,long permittedPoints){
    long newPoints = map << mask;
    if((newPoints & ~permittedPoints) != 0)
      return null;
    Cube newCube = (Cube)clone();
    newCube.map |= newPoints;
    newCube.mustBeZero &= ~mask;
    	newCube.dims = dims+1;
    //System.out.println("Generated cube "+newCube.toString());
    return newCube;
  }
  protected Cube expandDown(int mask, long permittedPoints){
    long newPoints = map >>> mask;
    if((newPoints & ~permittedPoints) != 0)
      return null;
    Cube newCube = (Cube)clone();
    newCube.map |= newPoints;
    newCube.mustBeOne &= ~mask;
    	newCube.dims = dims+1;
    //System.out.println("Generated cube "+newCube.toString());
    return newCube;
  }
	/** Find out whether this cube extends in a specified dimension.
	@param dim the dimension to probe.
	@return true if the cube has depth in that dimension, the result
	of having been extended in that dimension, false if not.
	*/
	boolean hasDepth(int dim){
		int mask = (1 << dim);
		if(((mustBeZero & mask) | (mustBeOne & mask)) != 0)
			return true;
		else
			return false;
	}
	/** Find out the location of this cube in a specified dimension.
	@param dim the dimension to probe.
	@return 0 if index dim of all points in the cube is 0, 1 if
	index dim of all points in the cube is 1.
	@throw SimplifierError if the cube has depth in the specified
	dimension, so its has points with both values.
	*/
	int getLoc(int dim)throws SimplifierError{
		int mask = (1 << dim);
		if((mustBeZero & mask) != 0)
			return 0;
		else if((mustBeOne & mask) != 0)
			return 1;
		else
			throw new SimplifierError("getLoc on expanded dimension "+dim);
	}
	/** Clone a Cube.
	@return Cube similar to this one.*/
	public Object clone(){
		try{
		  return super.clone();
		}catch(CloneNotSupportedException e){
			throw new SimplifierError("Cube clone not supported");
		}
	}
	
	/** Get dimensionality of this cube. Needed because
	solutions with equal numbers of cubes are selected based
	on prefering high dimension cubes.
	@return The number of dimensions in which this cube has
	been extended.
	*/
	public int getDims(){
		return dims;
	}
	
	/** Get a bit map of the points covered by this cube. Used
	in testing whether all required points have been covered.
	@return A bit map with bit n on if the point corresponding
	to minterm n is in the cube.
	*/
	public long getMap(){
		return map;
	}
  
  public String toString(){
    return "Map "+Long.toHexString(map)+" mustBeZero "+Integer.toHexString(mustBeZero)+
      " mustBeOne "+Integer.toHexString(mustBeOne);
  }
}