import java.util.Vector;

/** A CubeStack is a push/pop list of cubes, with methods
for printing the expression represented by the list of cubes,
and for evaluating whether one CubeStack represents a better
simplification than another.
*/
class CubeStack implements Cloneable{
	protected Vector stack = new Vector();  // The actual stack of cubes
	
	/** Push a cube onto the stack.
	@param c the Cube to push.
	*/
	void push(Cube c){
		stack.addElement(c);
	}
	
	/** Pop the most recently added cube that is still in
	the stack.
	@return the current top-of-stack Cube, which gets removed.
	*/
	Cube pop(){
		Cube c = (Cube)stack.lastElement();
		stack.removeElementAt(stack.size()-1);
		return c;
	}
	
	/** String showing the expression represented by the stack, 
	given a list of variable names.
	@param names The list of names of variables.
	@return A string representation of the list.
	*/
	String toString(String[] names){
		if(stack.size() == 0)
			return "0";
		StringBuffer result = new StringBuffer();
		for(int i=0;i<stack.size();i++){
			Cube c = (Cube)stack.elementAt(i);
			if(i>0)
				result.append("+");
     if(c.getDims() == names.length)
       result.append("1");
			else{
			  for(int n=0;n<names.length;n++){
			  	int m = names.length-1-n;
			  	if(c.hasDepth(m)){
					  result.append(names[n]);
					  if(c.getLoc(m)==0){
						  result.append("'");
					  }
					}
				}
			}
		}
		return result.toString();
	}
	
	public Object clone(){
		CubeStack cs;
		try{
		  cs = (CubeStack)super.clone();
		  cs.stack = (Vector)stack.clone();
		}catch(CloneNotSupportedException e){
			throw new SimplifierError("Clone problem: "+e.toString());
		}
		return cs;
	}
	
	/* Compare solution quality. One solution is better than another
	if it has fewer cubes, or the same number of cubes but with bigger
	cubes.
	@param o The solution to compare to.
	@return true iff this is better than o.
	*/
	public boolean betterThan(CubeStack o){
		if(stack.size() < o.stack.size())
			return true;
		if(stack.size() > o.stack.size())
			return false;
		int thisTot=0, thatTot=0;
		for(int i=0;i<stack.size();i++){
			thisTot += ((Cube)stack.elementAt(i)).getDims();
			thatTot += ((Cube)o.stack.elementAt(i)).getDims();
		}
		if(thisTot > thatTot)
			return true;
		else 
			return false;
	}
	
	/* Size of the stack.
	@return the number of cubes currently on the stack.
	*/
	public int size(){
		return stack.size();
	}
	
	/* Get a specific element of the stack.
	@param i index of required Cube.
	@return Cube i, where the first one to have
	been pused is considered Cube 0.
	*/
	protected Cube elementAt(int i){
		return (Cube)stack.elementAt(i);
	}
	
	/** Get a bit map of the points covered by this stack.
  @return A bit map with bit n on if the point corresponding
  to minterm n is in any the cube in the stack.
  */
  public long getMap(){
  	long map = 0;
  	for(int i=0;i<stack.size();i++)
  		map |= elementAt(i).getMap();
    return map;
  }

}