import java.util.*;
/** The actual simplifier class.*/
public class Simplifier{
	/* List of names with the name of the least significant
	term first.*/
	String[] names;
	/* Number of variables.*/
	int varCount;
	/* Maximum value of a valid minterm.*/
	int maxmin;
	/** Construct a new Simplifier that will use the
	specified series of names.
	@param names The names of the terms, in the same order
	as they would appear at the top of a truth table, most
	significant first. There must be no more than 6 names,
	because 64 bit longs are used to represent sets of bits
	in a space with 2^n elements.
	*/
	public Simplifier(String[] names){
		this.names = names;
		varCount = names.length;
		if(varCount > 6){
			throw new SimplifierError("More than 6 variables: "+varCount);
		}
    if(varCount == 0)
      maxmin = 1;
    else
		  maxmin = (1 << (varCount-1)) << 1;
		maxmin--;
	}
	
	/** The best solution so far. Initialized with the
	minterm expression.
	*/
	protected CubeStack best;
	/** Simplify an expression.
	@param minterms A list of integers each representing a minterm that must
  be covered by the solution.
  @param optional_minterms A list of integers each representing a minterm
  that may optionally be covered by the solution.
	@return String representation of the simplified expression.
	*/
	public String simplify(int[] minterms,int[] optional_minterms){
    for(int i=0;i<minterms.length;i++){
      if(minterms[i] > maxmin){
        throw new SimplifierError("minterm too large for names list");
      }
    }
    for(int i=0;i<optional_minterms.length;i++){
      if(optional_minterms[i] > maxmin){
        throw new SimplifierError("minterm too large for names list");
      }
    }
		/* Build an initial best solution, using the minterms as the
		basis. Note that a list of cubes representing each minterm
		as a single point is a valid solution, though unlikely to be
		optimal.*/
		best = new CubeStack();
		for(int i=0;i<minterms.length;i++){
			best.push(new Cube(minterms[i]));
		}
		/* Get ready to do recursive improvement. Need an initial
		trial cube, and sets of permitted and required points.*/
		if(minterms.length > 0){
		  Cube try1 = new Cube(minterms[0]);
			long permitted = best.getMap();
			long required = permitted & ~try1.getMap();
       /* Add optional minterms to the permitted set.*/
       for(int i=0;i<optional_minterms.length;i++){
         permitted |= (new Cube(optional_minterms[i])).getMap();
       }
			if(required != 0)
		    improve(new CubeStack(),try1,permitted,required,0);
		}
		/* Return a string representing the best solution.*/
		return best.toString(names);
	}
	
	/** The key method for the simplification. Examines a set
	of possible solutions to see if there is a better one than
	the current best. The current state of the simplification
	includes a list of cubes that have already been chose, a
	cube that is under construction, and a set of points that
	must be covered by the remainder of the solution.
	@param inStack A stack of cubes that are considered to be
	in the solution being examined.
	@param inCube The cube that is currently being expanded with
	a view to pushing it onto the stack.
	@param inPermit The map of points that are permitted to be
	covered by any solution.
	@param inReq The map of points that are required to be covered
	by expansion of the current cube and addition of successors. It
	is all points that are required to be covered by a valid solution
	and that are not covered by the cubes in inStack.
	@param fixedDim The lowest numbered fixedDim dimensions of inCube
	are considered to be bound to their current values. This instance
	of improve must expand inStack only in dimensions fixedDim and up.
	*/
	void improve(CubeStack inStack,Cube inCube, long permitted, 
		long inReq, int fixedDim)
  {
  //System.out.println("improve called cube "+inCube.toString()+
  // " req "+Long.toHexString(inReq));
  	/* First try to expand the current cube. If it can be expanded,
  	look both for solutions based on the expanded version and based on
  	not expanding that dimension but looking for possible expansion in
  	later dimensions.
  	*/
  	for(int i=fixedDim;i<varCount;i++){
  		Cube nextCube = inCube.expand(i,permitted);
  		if(nextCube != null){
  			long subReq = inReq & ~nextCube.getMap();
  			improve(inStack,nextCube,permitted,subReq,i+1);
  			improve(inStack,inCube,permitted,inReq,i+1);
  			return;
  		}
  	}
  	/* There were no more dimensions in which the current cube can
  	be expanded. Push it on the stack and see if it is possible to
  	form a new, improved solution based on it.*/
  	inStack.push(inCube);
  	/* Once the push has been done it is essential to make sure that
  	the pop is also done, so use a try-finally block.*/
  	try{
  	  if(inReq == 0){
  		  /* There are no more points that need to be covered. See
  		  whether this solution is better than the best so far.*/
  		  if(inStack.betterThan(best)){
  			  best = (CubeStack)inStack.clone();
  		  }
  	  }else{
  		  /* More points need to be covered, so this is not yet a solution.
  		  If it contains fewer cubes than the current best it may be the
  		  basis of a winner with more cubes added. If it already contains
  		  at least as many cubes as the current solution it cannot be the
  		  basis of a better solution so this line should be abandoned.*/
  		  if(inStack.size() < best.size()){
  			  /* select the next point to add.*/
  			  for(int i=0;;i++){
  				  if( ( inReq & (1L << i) ) != 0){
  				  	/* Continue the attempted solution using this as a basis,
  				  	making a single point from inReq the new current cube. This
  				  	call to improve will lead to an attempt to first expand it,
  				  	and then add other cubes if necessary.*/
  				  	improve(inStack,new Cube(i),permitted,inReq,0);
  				  	break;
  				  }
  			  }
  		  }
  	  }
  	}finally{
  		inStack.pop();
  	}
	}
}