import java.awt.*;

/** Applet to simplify a truth table, generating expressions.*/
public class SimplifyApplet extends java.applet.Applet{
  
  /* Area where the truth table will be entered.*/
  protected TextArea input;
  /* Area where results or error messages will be displayed.*/
  protected TextArea output;
  /* Button pushed to indicate that input is available.*/
  protected Button compute;
  
  /** Set up display.*/
  public void init()
  {
    /* (Re)initialize class variables.*/
    input = new TextArea();
    output = new TextArea();
    output.setEditable(false);
    compute = new Button("Compute");
    
    /* Lay out the display.*/
    GridBagLayout gb = new GridBagLayout();
    setLayout(gb);
    GridBagConstraints c = new GridBagConstraints();
    Label lab = new Label("Input");
    add(lab);
    c.gridx = 0;
    c.gridy = 0;
    gb.setConstraints(lab,c);
    lab = new Label("Results");
    add(lab);
    c.gridx = 1;
    gb.setConstraints(lab,c);
    add(input);
    c.fill = GridBagConstraints.BOTH;
    c.gridx = 0;
    c.gridy = 1;
    c.weightx = 10;
    c.weighty = 10;
    c.gridheight = GridBagConstraints.REMAINDER;
    gb.setConstraints(input,c);
    
    add(output);
    c.gridx = 1;
    c.weightx = 7;
    c.gridheight = 1;
    gb.setConstraints(output,c);
    
    add(compute);
    c.gridx = 1;
    c.gridy = 2;
    c.weightx = 0;
    c.weighty = 0;
    c.fill = GridBagConstraints.NONE;
    gb.setConstraints(compute,c);
  }
  
  /** Do the actual computation.
  @param truthTable A string representation of the input truth table.
  @return A string representation of the simplified forms of the dependent
  varable functions and their inverses.*/
  protected String doCompute(String truthTable){
    /* Create a convenient representation of the truth table. This
    will cause an exception of there are errors in the format of
    truthTable.*/
    try{
      TruthTable tt = new TruthTable(truthTable);
      /* Get the independent variable names.*/
      String[] indep = tt.getIndepVars();
      /* Create a simplifier to handle these variables.*/
      Simplifier simp = new Simplifier(indep);
      /* Create a function parameter list based on these
      variables.*/
      String paramList = toParamList(indep);
      /* Accumulate results into a string buffer.*/
      StringBuffer result = new StringBuffer();
      String[] depVars = tt.getDepVars();
      for(int i=0;i<depVars.length;i++){
        int[] required = tt.getRequired(i);
        int[] optional = tt.getOptional(i);
        /* Do the forwards function.*/
        result.append(depVars[i]);
        result.append(paramList);
        result.append("=");
        result.append(simp.simplify(required,optional));
        result.append("\n");
        /* Do its inverse.*/
        result.append(depVars[i]);
        result.append("'");
        result.append(paramList);
        result.append("=");
        required = tt.getProhibited(i);
        result.append(simp.simplify(tt.getProhibited(i),tt.getOptional(i)));
        result.append("\n");
      }
      return result.toString();
    }catch(SimplifierError e){
      return "ERROR: "+e.getMessage();
    }
  }
  
  /** Convert an array of variables names to a parameter list.
  @param names An array of parameter names.
  @return A parenthesis enclosed representation of the names as
  a parameter list.*/
  String toParamList(String[] names){
    StringBuffer result = new StringBuffer();
    result.append("(");
    for(int i=0;i<names.length;i++){
      if(i>0)
        result.append(",");
      result.append(names[i]);
    }
    result.append(")");
    return result.toString();
  }
  
  /** React to the button being pressed.*/
  public boolean action(Event evt, Object what)
  {
    if(evt.target == compute){
      /** Got a compute event.*/
      output.setText("Computing");
      output.setText(doCompute(input.getText()));
      repaint();
      return true;
    }
    return super.action(evt,what);
  }
}