import java.io.EOFException;
import java.io.IOException;
import java.io.LineNumberReader;
import java.io.StringReader;
import java.util.StringTokenizer;
import java.util.Vector;

/** Representation of a truth table.*/
class TruthTable{
  protected int indepVars;  // Number of independent variables.
  protected String[] indepVarNames;
  protected int depVars;  // Number of dependent variables.
  protected String[] depVarNames;
  protected int allVars;  // Total number of variables.
  /* Bits will be represented by values 0, 1, and X, where
  X indicated that it can either by 0 or 1.*/
  protected static final int X=-1;
  int[][] allData;
  int dataLineCount;
  
  /** Create a new truth table based on a string representation.
  @param table The truth table in conventional format, with rows separated by
  cr, nl, or their combination and elements in rows separated by whitespace.
  The first row must contain a variable name for each variables. The remainder
  of the table must contain 2^N data rows, where N is the number of independent
  variables. It is assumed to be a full truth table in conventional order. The
  first N fields in the header row will be the names of the independent variables,
  the remaining names will be used as the dependent variable names.
  @throw SimplifierError Some validation error processing the table string. The
  exception message explains the error.
  */
  TruthTable(String table) throws SimplifierError{
    /* Begin by splitting on newlines.*/
    LineNumberReader in = new LineNumberReader(new StringReader(table));
    /* Get the data into a Vector.*/
    Vector lines = new Vector();
    boolean done = false;
    while(!done)try{
      String s = in.readLine();
      if(s == null)
        done = true;
      else
        lines.addElement(s);
    }catch(IOException e){
      throw new SimplifierError(e.getMessage());
    }
    /* Check the number of lines - there should be one header line
    and 2^N detail lines where there are N independent variables.*/
    dataLineCount = lines.size()-1;
    indepVars = -1;
    if(dataLineCount <= 0)
      throw new SimplifierError("No data in truth table.");
    if(dataLineCount > 64)
      throw new SimplifierError("Truth table too large.");
    for(int i=0;i<=6;i++){
      if(dataLineCount == (1 << i)){
        indepVars = i;
        break;
      }
    }
    if(indepVars == -1){
      throw new SimplifierError(
        "Truth table must contain 2^N data lines");
    }
    /* Now process the header line.*/
    StringTokenizer st;
    st = new StringTokenizer((String)lines.elementAt(0));
    depVars = st.countTokens()-indepVars;
    if(depVars <= 0)
      throw new SimplifierError("Too few columns for number of rows");
    allVars = depVars+indepVars;
    indepVarNames = new String[indepVars];
    for(int i=0;i<indepVars;i++)
      indepVarNames[i] = st.nextToken();
    depVarNames = new String[depVars];
    for(int i=0;i<depVars;i++)
      depVarNames[i] = st.nextToken();
    /* Now process the data rows, converting each to an array
    of integers.*/
    allData = new int[dataLineCount][];
    for(int row=0;row<dataLineCount;row++){
      allData[row] = new int[allVars];
      String line = (String)lines.elementAt(row+1);
      StringTokenizer stl = new StringTokenizer(line);
      if(stl.countTokens() != allVars)
        throw new SimplifierError("Row "+row+" has wrong number of columns.");
      for(int col=0;col<allVars;col++){
        String s = stl.nextToken();
        if(s == null)
          throw new SimplifierError("Missing data element");
        if(s.equals("0")){
          allData[row][col]=0;
        }else if(s.equals("1")){
          allData[row][col]=1;
        }else if(col >= indepVars && s.equalsIgnoreCase("x")){
          /* Don't care data is allowed only in dependent variable columns.*/
          allData[row][col]=X;
        }else{
          throw new SimplifierError("Bad data element row "+row+" column "+col+
              " value "+s);
        }       
      }
    }
    /* allData now contains the data elements in a simple integer array.*/
  }
  
  /** Get the names of the independent variables.
  @return the independent variables names in order of appearance.
  */
  String[] getIndepVars(){
    return indepVarNames;
  }
  
  /** Get the names of the dependent variables.
  @return the dependent variable names in order of appearance.
  */
  String[] getDepVars(){
    return depVarNames;
  }
  
  /** Get the required minterm list for a specified dependent variable.
  @param varNo the variable number, must be between 0 and getDepVars.length-1.
  @return the minterm numbers that must be covered by a valid simplification
  of dependent variable varNo.
  */
  int[] getRequired(int varNo){
    setUpVar(varNo);
    return requiredTerms;
  }
  
  /** Get the optional minterm list for a specified dependent variable.
  @param varNo the variable number, must be between 0 and getDepVars.length-1.
  @return the minterm numbers that may optionally be covered by a valid 
  simplification of dependent variable varNo. Note that a variable and its
  inverse use the same optional list, since each optional minterm is neither
  required nor prohibited for each of them.
  */
  int[] getOptional(int varNo){
    setUpVar(varNo);
    return optionalTerms;
  }
  
  /** Get the prohibited minterm list for a specified dependent variable.
  @param varNo the variable number, must be between 0 and getDepVars.length-1.
  @return the minterm numbers that must not be covered by a valid simplification
  of dependent variable varNo. Note that the prohibited list is also the required
  list for the inverse of the dependent variable.
  */
  int[] getProhibited(int varNo){
    setUpVar(varNo);
    return prohibitedTerms;
  }
  
  /** Set up common information for a specified dependent variable.
  @param varNo index of the required variable.*/
  protected void setUpVar(int varNo){
    if(varNo == lastVarNo)
      return;
    lastVarNo = varNo;
    int wkVar = varNo+indepVars;
    /* First classify and count them.*/
    prohibitedCount = 0;
    requiredCount = 0;
    for(int i=0;i<dataLineCount;i++){
      if(allData[i][wkVar] == 0)
        prohibitedCount++;
      else if(allData[i][wkVar] == 1)
        requiredCount++;
    }
    /* Set up the correct array sizes.*/
    requiredTerms = new int[requiredCount];
    prohibitedTerms = new int[prohibitedCount];
    optionalTerms = new int[dataLineCount-(requiredCount+prohibitedCount)];
    /* Store the terms in the appropriate arrays.*/
    int rt = 0;
    int pt = 0;
    int ot = 0;
    for(int i=0;i<dataLineCount;i++){
      switch(allData[i][wkVar]){
        case 0:
          prohibitedTerms[pt] = i;
          pt++;
          break;
        case 1:
          requiredTerms[rt] = i;
          rt++;
          break;
        case X:
          optionalTerms[ot] = i;
          ot++;
          break;
      }
    }
  }
  
  /* The variables controlled by setUpVar.*/
  int lastVarNo = -1;
  int requiredCount;
  int prohibitedCount;
  int[] requiredTerms;
  int[] optionalTerms;
  int[] prohibitedTerms;
}